(*
  Atari3 - read/write Atari disks on the PC

  Copyright (c) 1992-1995 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE Atari3;

IMPORT SYSTEM,Lib,IO,FIO;

VAR drive:SHORTCARD;
    Invert:BOOLEAN;
    drive_type:SHORTCARD; (* type of drive, as listed in CMOS *)

VAR FormatInterleaveAtari:BOOLEAN;


VAR SectorsPerTrack,SectorLength:CARDINAL;

VAR disk_param_table[0:78H]:ADDRESS;
    old_table:ADDRESS;
TYPE a11=ARRAY[0..10] OF BYTE;
TYPE a11p=POINTER TO a11;
VAR new_table:a11;

CONST MaxRetry=5;

TYPE Array18=ARRAY[0..17] OF SHORTCARD;

CONST InterleaveDD=Array18(1,12,4,15,7,18,10,2,13,5,16,8,11,3,14,6,17,9);

PROCEDURE FormatTrack(t:CARDINAL);
VAR trkinfo:ARRAY[0..25] OF RECORD
                               track,side,secno,byte:SHORTCARD
                            END;
    i:CARDINAL;
    r:SYSTEM.Registers;
    retry:CARDINAL;
BEGIN
    FOR i:=0 TO SectorsPerTrack-1 DO
        trkinfo[i].track:=SHORTCARD(t);
        trkinfo[i].side:=0;
        trkinfo[i].secno:=SHORTCARD(i+1);
        IF FormatInterleaveAtari THEN
           IF (SectorLength=256) & (SectorsPerTrack=18) THEN
              trkinfo[i].secno:=InterleaveDD[i]
           END;
        END;
        IF SectorLength=128 THEN
            trkinfo[i].byte:=0;
        ELSIF SectorLength=256 THEN
            trkinfo[i].byte:=1;
        ELSE
            trkinfo[i].byte:=2;
        END;
    END;
    retry:=0;
    REPEAT
        r.AH:=5;
        r.DL:=drive;
        r.DH:=0;
        r.CH:=SHORTCARD(t);
        r.CL:=0;
        r.AL:=SHORTCARD(SectorsPerTrack);
        r.ES:=SYSTEM.Seg(trkinfo);
        r.BX:=SYSTEM.Ofs(trkinfo);
        Lib.Intr(r,13H);
        ok := (r.Flags * {SYSTEM.CarryFlag}={});
        INC(retry);
    UNTIL ok OR (retry=MaxRetry);
END FormatTrack;


PROCEDURE FormatDisk;
VAR track:CARDINAL;
BEGIN
    track:=0;
    REPEAT
         FormatTrack(track);
         INC(track);
    UNTIL (track=40) OR (ok=FALSE)
END FormatDisk;

PROCEDURE SaveOldTable;
BEGIN
    old_table:=disk_param_table
END SaveOldTable;

PROCEDURE RestoreOldTable;
BEGIN
    disk_param_table:=old_table
END RestoreOldTable;

PROCEDURE SetNewTable;
VAR p:a11p;
BEGIN
    p:=a11p(old_table);
    new_table:=p^;
    new_table[3]:=1;
    new_table[4]:=3FH;
    new_table[5]:=0CH;
    new_table[7]:=1CH;
    new_table[8]:=0FFH;
(*

    new_table:=a11(0DFH,02H,025H,
                   01H,
                   3FH,
(*
                   2AH,080H,30H,0F6H,014H,08H);
                   1BH,0FFH,54H,0F6H,0FH,08H);
                   1FH,0FFH,24H,0FFH,0FH,08H);
                   05H,0FFH,10H,0FFH,0FH,08H);
*)
                   0CH,0FFH,1CH,0FFH,0FH,08H); (* 29,24*)
*)
    disk_param_table:=ADR(new_table);
END SetNewTable;

PROCEDURE InitController;
 VAR r:SYSTEM.Registers;
BEGIN
    r.AH:=0;
    r.DL:=drive;
    Lib.Intr(r,13H);
END InitController;


PROCEDURE WriteSector(sec:CARDINAL; buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    i:CARDINAL;
    s,t:SHORTCARD;
    retry:CARDINAL;
BEGIN
    IF Invert THEN
        FOR i:=0 TO SectorLength-1 DO
            buf[i]:=BYTE(255-SHORTCARD(buf[i]))
        END;
    END;

    DEC(sec);
    t:=SHORTCARD(sec DIV SectorsPerTrack);
    s:=SHORTCARD((sec MOD SectorsPerTrack)+1);

    retry:=0;
    REPEAT
        r.AH:=3;
        r.DL:=drive;
        r.DH:=0;
        r.CH:=t;
        r.CL:=s;
        r.AL:=1;
        r.ES:=SYSTEM.Seg(buf);
        r.BX:=SYSTEM.Ofs(buf);
        Lib.Intr(r,13H);
        ok := (r.Flags * {SYSTEM.CarryFlag}={});
        INC(retry);
    UNTIL ok OR (retry=MaxRetry);

END WriteSector;

PROCEDURE ReadSector(sec:CARDINAL; VAR buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    i:CARDINAL;
    s,t:SHORTCARD;
    retry:CARDINAL;
BEGIN
    retry:=0;

    DEC(sec);
    t:=SHORTCARD(sec DIV SectorsPerTrack);
    s:=SHORTCARD((sec MOD SectorsPerTrack)+1);

    REPEAT
        r.AH:=2;
        r.DL:=drive;
        r.DH:=0;
        r.CH:=t;
        r.CL:=s;
        r.AL:=1;
        r.ES:=SYSTEM.Seg(buf);
        r.BX:=SYSTEM.Ofs(buf);
        Lib.Intr(r,13H);
        ok := (r.Flags * {SYSTEM.CarryFlag}={});
        INC(retry);
    UNTIL ok OR (retry=MaxRetry);

    IF (ok & Invert) THEN
        FOR i:=0 TO SectorLength-1 DO
            buf[i]:=BYTE(255-SHORTCARD(buf[i]))
        END;
    END;
END ReadSector;

PROCEDURE Fix360;
(* Sets correct format-type for 360k-disk (in 360k/1.2m drive) *)
VAR r:SYSTEM.Registers;
BEGIN
    r.AH:=17H;
    IF (drive_type=1) THEN
       r.AL:=1;
    ELSE
       r.AL:=2;
    END;
    r.DL:=drive;
    Lib.Intr(r,13H);
END Fix360;


PROCEDURE InitSD;
BEGIN
    SectorLength:=128;
    SectorsPerTrack:=18;
    new_table[3]:=0;
    new_table[4]:=12H;
    Invert:=TRUE;
    InitController;
END InitSD;

PROCEDURE InitED;
BEGIN
    SectorLength:=128;
    SectorsPerTrack:=26;
    new_table[3]:=0;
    new_table[4]:=1AH;
    Invert:=TRUE;
    InitController;
END InitED;

PROCEDURE InitDD;
BEGIN
    SectorLength:=256;
    SectorsPerTrack:=18;
    new_table[3]:=1;
    new_table[4]:=12H;
    Invert:=TRUE;
    InitController;
    Fix360;
END InitDD;


PROCEDURE InitPCDD;
BEGIN
    SectorLength:=512;
    SectorsPerTrack:=9;
    new_table[3]:=2;
    new_table[4]:=0FH;
    Invert:=FALSE;
    InitController;
END InitPCDD;

PROCEDURE CheckDriveType;
VAR r:SYSTEM.Registers;
    p:a11p;
BEGIN
 r.AH:=08H;
 r.DL:=drive;
 Lib.Intr(r,13H);
 drive_type:=SHORTCARD(BITSET(r.BL)*BITSET(15));
 p:=a11p([r.ES:r.DI]);
END CheckDriveType;

PROCEDURE SetDriveA;
BEGIN
 drive:=0;
 CheckDriveType;
END SetDriveA;

PROCEDURE SetDriveB;
BEGIN
 drive:=1;
 CheckDriveType;
END SetDriveB;

PROCEDURE SetInterleaveAtari;
BEGIN
    FormatInterleaveAtari:=TRUE
END SetInterleaveAtari;

PROCEDURE SetInterleave1;
BEGIN
    FormatInterleaveAtari:=FALSE
END SetInterleave1;

PROCEDURE ClearDirDD;
VAR sec:ARRAY[0..255] OF BYTE;
	i:CARDINAL;
BEGIN
	Lib.Fill(ADR(sec),256,0);
	sec[0]:=2;
	sec[1]:=195;
	sec[2]:=2;
	sec[3]:=195;
	sec[4]:=2;
	sec[10]:=15;
	FOR i:=11 TO 54 DO
        sec[i]:=255;
	END;
	sec[56]:=127;
	FOR i:=57 TO 99 DO
	    sec[i]:=255;
	END;
    WriteSector(360,sec);
	Lib.Fill(ADR(sec),256,0);
    i:=361;
    WHILE ok & (i<=368) DO
        WriteSector(i,sec);
        INC(i);
    END;	
END ClearDirDD;



VAR OldCont:PROC;

PROCEDURE Terminate;
BEGIN
    RestoreOldTable;
    OldCont;
END Terminate;

PROCEDURE InitAtari3;
BEGIN
	SetDriveA;
    SaveOldTable;
    SetNewTable;
    SetInterleaveAtari;
    InitDD;
    Lib.Terminate(Terminate,OldCont);
END InitAtari3;

END Atari3.
