(*
  hiaslib - various helper functions

  Copyright (c) 1990-1994 by Matthias Reichl <hias@horus.com>
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE HiasLib;

IMPORT IO,Lib,SYSTEM,Graph,Storage,Str;

TYPE
  tinyint      = [0..7];
  bs           = SET OF tinyint;
  bp           = POINTER TO bs;

CONST Ord0=ORD('0');

VAR bitmap [0A000H:0000H]:ARRAY [0..65535] OF BYTE;
    Screen16 [0A000H:0] : ARRAY[0..0] OF bs;

VAR ProgName:ARRAY[0..255] OF CHAR;

PROCEDURE PlotMCGA(x,y,color:CARDINAL);

BEGIN
 bitmap[x+(y<<8)+(y<<6)]:=BYTE(color);
END PlotMCGA;

PROCEDURE PlotBios(x,y,color:CARDINAL);
VAR r:SYSTEM.Registers;

BEGIN
 r.AH:=0CH;
 r.AL:=BYTE(color);
 r.BH:=0;
 r.CX:=x;
 r.DX:=y;
 Lib.Intr(r,10H);
END PlotBios;

PROCEDURE SetColor(Register:CARDINAL;Red,Green,Blue:SHORTCARD);
VAR r:SYSTEM.Registers;

BEGIN
 r.AH:=10H;
 r.AL:=10H;
 r.BX:=Register;
 r.DH:=Red;
 r.CH:=Green;
 r.CL:=Blue;
 Lib.Intr(r,10H)
END SetColor;

PROCEDURE SetPalette(FirstRegister,Number:CARDINAL;
                     VAR ColorTable:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;

BEGIN
 r.AH:=10H;
 r.AL:=12H;
 r.ES:=SYSTEM.Seg(ColorTable);
 r.DX:=SYSTEM.Ofs(ColorTable);
 r.BX:=FirstRegister;
 r.CX:=Number;
 Lib.Intr(r,10H);
END SetPalette;

PROCEDURE SetPalDirect(FirstRegister,Number:CARDINAL;
                       VAR ColorTable:ARRAY OF BYTE);
VAR i:CARDINAL;

BEGIN
 SYSTEM.DI;
 SYSTEM.Out(3C8H,BYTE(FirstRegister));
 FOR i:=0 TO Number*3-1 DO
  SYSTEM.Out(3C9H,ColorTable[i])
 END;
 SYSTEM.EI
END SetPalDirect;

PROCEDURE ReadRegister(Register:SHORTCARD; VAR r,g,b:SHORTCARD);
BEGIN
 SYSTEM.Out(03C7H,Register);
 r:=SYSTEM.In(03C9H);
 g:=SYSTEM.In(03C9H);
 b:=SYSTEM.In(03C9H);
END ReadRegister;


PROCEDURE GetInformation(VAR Buffer:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
BEGIN
 r.AH:=1BH;
 r.BX:=0;
 r.ES:=SYSTEM.Seg(Buffer);
 r.DI:=SYSTEM.Ofs(Buffer);
 Lib.Intr(r,10H);
END GetInformation;


PROCEDURE SetTimer(time:LONGCARD);
VAR r:SYSTEM.Registers;
BEGIN
 r.AH:=1;
 r.CX:=CARDINAL(time DIV 65536);
 r.DX:=CARDINAL(time MOD 65536);
 Lib.Intr(r,1AH);
END SetTimer;

PROCEDURE ReadTimer(VAR time:LONGCARD);
VAR r:SYSTEM.Registers;
BEGIN
 r.AH:=0;
 Lib.Intr(r,1AH);
 time:=LONGCARD(r.DX)+65536*LONGCARD(r.CX);
END ReadTimer;

PROCEDURE ReadClock(VAR time:ARRAY OF CHAR);
VAR r:SYSTEM.Registers;
    temp:CARDINAL;
BEGIN
 r.AH:=2;
 Lib.Intr(r,1AH);
 time[0]:=CHR(Ord0+(r.CX DIV 4096));
 time[1]:=CHR(Ord0+((r.CX DIV 256) MOD 16));
 time[2]:=':';
 time[3]:=CHR(Ord0+((r.CX DIV 16) MOD 16));
 time[4]:=CHR(Ord0+(r.CX MOD 16));
 time[5]:=':';
 time[6]:=CHR(Ord0+(CARDINAL(r.DH) DIV 16));
 time[7]:=CHR(Ord0+(CARDINAL(r.DH) MOD 16));
 time[8]:=CHR(0);
END ReadClock;

PROCEDURE VideoMode(mode:CARDINAL);
VAR r:SYSTEM.Registers;
BEGIN
 r.AH:=0;
 r.AL:=SHORTCARD(mode);
 Lib.Intr(r,10H);
END VideoMode;

PROCEDURE CursorSize(StartLine,EndLine:CARDINAL);
VAR r:SYSTEM.Registers;
BEGIN
 r.AH:=1;
 r.CH:=SHORTCARD(StartLine);
 r.CL:=SHORTCARD(EndLine);
 Lib.Intr(r,10H);
END CursorSize;

PROCEDURE Exec ( command : ARRAY OF CHAR ;
                 params  : ARRAY OF CHAR;
                 foo: ADDRESS ) :CARDINAL;

(*
   Executes a named program, using all free heap memory to do so.
   All memory allocation/deallocation is automatic.
   The program name should include the extension and full path

     eg
         Exec('\COM\ED.EXE','AFILE') ;

   The string params will be passed to the program as the
   'command line'.

*)
VAR
   memptr  : ADDRESS ;
   memsize : CARDINAL ;
   reply   : CARDINAL ;
BEGIN
   memsize := Storage.HeapAvail ( Storage.MainHeap )-8 ; (* 8 for MSDOS relocation *)
   Storage.HeapAllocate(Storage.MainHeap,memptr,memsize) ;
   reply := Lib.Execute(command,params,memptr,memsize) ;
   Storage.HeapDeallocate(Storage.MainHeap,memptr,memsize) ;
   RETURN reply;
END Exec ;

PROCEDURE GetEnv( env_var:ARRAY OF CHAR;
                  VAR val:ARRAY OF CHAR;
                  default:ARRAY OF CHAR);
VAR i,l,p:CARDINAL;
    s:Lib.CommandType;
BEGIN
    i:=0;
    Str.Caps(env_var);
    l:=Str.Length(env_var);
    REPEAT
         s:=Lib.Environment(i);
         IF (s#NIL) & (s^[0]#0C) THEN
            p:=Str.Pos(s^,env_var);
            IF (p=0) & (s^[l]="=") THEN
               Str.Slice(val,s^,l+1,1024);
               RETURN
            END;
         END;
         INC(i);
    UNTIL (s=NIL) OR (s^[0]=0C);
    Str.Copy(val,default);
END GetEnv;

PROCEDURE GetProgName(VAR n:ARRAY OF CHAR);
VAR adr,i:CARDINAL;
TYPE pta=POINTER TO ARRAY[0..1024] OF CHAR;
     ptb=POINTER TO BYTE;
     ptc=POINTER TO CARDINAL;
VAR env:CARDINAL;
    e:pta;
BEGIN
    env:=[Lib.PSP:2CH ptc]^;
    adr:=0;
    LOOP
       i:=Lib.ScanR([env:adr ptb],65535-i,0);
       IF i=0 THEN EXIT END;
       adr:=adr+i+1;
    END;
    e:=[env:adr+3 pta];
    Str.Copy(n,e^);
END GetProgName;

PROCEDURE GetPathFromFilename(fn:ARRAY OF CHAR;
                              VAR path:ARRAY OF CHAR);
VAR i,j,l:CARDINAL;
BEGIN
    l:=Str.Length(fn);
    IF l=0 THEN
       path[0]:=0C;
       RETURN
    END;

    i:=Lib.ScanL(ADR(fn[l-1]),l-1,"\");
    IF i=l-1 THEN
       path[0]:=0C
    ELSE
       Str.Slice(path,fn,0,l-i);
    END;
END GetPathFromFilename;

BEGIN
 Plot:=PlotBios;
END HiasLib.
