(*
  Input - extended line input with edit

  Copyright (c) 1991-1993 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE Input;

IMPORT IO,Str;

	VAR KeyPushed:BOOLEAN;
		MerkKey:KeyType;

	TYPE CharSet=SET OF CHAR;

    TYPE LONGSTR=ARRAY[0..MaxStringLen] OF CHAR;
    (* ------------------------------------------------------------------ *)
    (* String mit maximaler Lnge (wie in Common definiert)               *)
    (* ------------------------------------------------------------------ *)

    CONST Alpha=CharSet{'a'..'z','A'..'Z','','','','','','','','0'..'9'};
    (* ------------------------------------------------------------------ *)
    (* Gltige Buchstaben                                                 *)
    (* ------------------------------------------------------------------ *)

	CONST eos=0C;

	PROCEDURE PushKey(k:KeyType);
	BEGIN
		KeyPushed:=TRUE;
		MerkKey:=k
	END PushKey;

	PROCEDURE TestKeyPushed():BOOLEAN;
	BEGIN
		RETURN KeyPushed;
	END TestKeyPushed;

	PROCEDURE PopKey():KeyType;
	BEGIN
		KeyPushed:=FALSE;
		RETURN MerkKey;
	END PopKey;

	PROCEDURE ClearPushBuffer;
	BEGIN
		KeyPushed:=FALSE;
	END ClearPushBuffer;

    PROCEDURE RdKeyExt():KeyType;
    (* ------------------------------------------------------------------ *)
    (* Erweitertes ReadKey - untersttzt auch Funktionstasten             *)
    (* ------------------------------------------------------------------ *)
    (* PARAMETER:                                                         *)
    (*     NAME          BESCHREIBUNG:                                    *)
    (*                                                                    *)
    (* RETURNWERT:                                                        *)
    (*                   Erweiterte 'Taste'                               *)
    (* ------------------------------------------------------------------ *)
    (* HISTORISCHE INFORMATION:                                           *)
    (*     NDERUNGEN:                                                    *)
    (* ------------------------------------------------------------------ *)
    (* ERGNZUNGEN:                                                       *)
    (*                                                                    *)
    (* ------------------------------------------------------------------ *)

    VAR k:CHAR;
        key:KeyType;
    BEGIN
        key.level:=normal; key.key:=0C;  (* =keine Taste. Das Zeichen mit *)
                                   (* dem ASCII-Code 0 kann am PC mit der *)
                                        (* Tastatur nicht erzeugt werden. *)

                         (* Bemerkung: auf dem PC wird beim Drcken einer *)
                            (* Funktions-Taste zuerst ein Character 0 und *)
                           (*dann der eigentliche F-Tasten-Code gesendet. *)
        REPEAT
            k:=IO.RdKey();
            IF ORD(k)=0 THEN                            (* Funktionstaste *)
                key.level:=function;
                k:=IO.RdKey();       (* nchster Wert ist Art der F-Taste *)
                CASE ORD(k) OF
				  15: key.fkey:=tab2;
                | 59..68: key.fkey:=VAL(FKeyType,ORD(k)-59+ORD(F1));
                | 84..93: key.fkey:=VAL(FKeyType,ORD(k)-84+ORD(SF1));
                | 71: key.fkey:=pos1;
                | 72: key.fkey:=up;
                | 73: key.fkey:=pageup;
                | 75: key.fkey:=left;
                | 77: key.fkey:=right;
                | 79: key.fkey:=ende;
                | 80: key.fkey:=down;
                | 81: key.fkey:=pagedown;
                | 82: key.fkey:=einf;
                | 83: key.fkey:=entf;
                | 115: key.fkey:=left2;
                | 116: key.fkey:=right2;
                ELSE
                  key.level:=normal;   (* ungltigt F-Taste -> ignorieren *)
                  key.key:=CHR(32);
                END;
            ELSE
                CASE ORD(k) OF
                  32..126,128..254: key.level:=normal; key.key:=k;
                                           (* darstellbares ASCII-Zeichen *)

                     (* nicht darstellbare ASCII-Zeichen = 'Sondertasten' *)
                | 27: key.level:=function; key.fkey:=esc;
                | 13: key.level:=function; key.fkey:=return;
                |  8: key.level:=function; key.fkey:=delete;
                |  9: key.level:=function; key.fkey:=tab;
                | 25: key.level:=function; key.fkey:=CtrlY;
                END;
            END;
        UNTIL NOT((key.level=normal) & (key.key=CHR(0)));
        RETURN key;
    END RdKeyExt;

    PROCEDURE WriteString(str:ARRAY OF CHAR;length:CARDINAL);
    (* ------------------------------------------------------------------ *)
    (* gib einen String aus und flle den Rest  ( length-LEN(str) )  mit  *)
    (* Blanks                                                             *)
    (* ------------------------------------------------------------------ *)
    (* PARAMETER:                                                         *)
    (*     NAME          BESCHREIBUNG:                                    *)
    (*     str           String                                           *)
    (*     length        Lnge, bis zu den mit Blanks aufgefllt wird     *)
    (* ------------------------------------------------------------------ *)
    (* HISTORISCHE INFORMATION:                                           *)
    (*     NDERUNGEN:                                                    *)
    (* ------------------------------------------------------------------ *)
    (* ERGNZUNGEN:                                                       *)
    (* ------------------------------------------------------------------ *)

    VAR i:CARDINAL;
    BEGIN
		Window.CursorOff;
		IO.WrStrAdj(str,-1-INTEGER(length));
(*
        IO.WrStr(str);
		i:=Str.Length(str);
        IF length>=i THEN                    (* Rest mit Blanks auffllen *)
			IO.WrCharRep(" ",length-i+1);
        END;
*)
    END WriteString;

    PROCEDURE WriteStringRightJustified(str:ARRAY OF CHAR;length:CARDINAL);
    BEGIN
		Window.CursorOff;
		IO.WrStrAdj(str,length+1);
    END WriteStringRightJustified;

    PROCEDURE WriteStringXY(str:ARRAY OF CHAR;length:CARDINAL;x,y:CARDINAL);
    (* ------------------------------------------------------------------ *)
    (* gib einen String aus und flle den Rest  ( length-LEN(str) )  mit  *)
    (* Blanks                                                             *)
    (* ------------------------------------------------------------------ *)
    (* PARAMETER:                                                         *)
    (*     NAME          BESCHREIBUNG:                                    *)
    (*     str           String                                           *)
    (*     length        Lnge, bis zu den mit Blanks aufgefllt wird     *)
    (*     x,y           absolute Screen-Koordinaten                      *)
    (* ------------------------------------------------------------------ *)
    (* HISTORISCHE INFORMATION:                                           *)
    (*     NDERUNGEN:                                                    *)
    (* ------------------------------------------------------------------ *)
    (* ERGNZUNGEN:                                                       *)
    (*     keine Koordinaten-berprfung!                                 *)
    (* ------------------------------------------------------------------ *)

    VAR i:CARDINAL;
    BEGIN
		Window.CursorOff;
		Window.GotoXY(x,y);
        IO.WrStr(str);
		i:=Str.Length(str);
        IF length>=i THEN                    (* Rest mit Blanks auffllen *)
			IO.WrCharRep(" ",length-i+1);
        END;
    END WriteStringXY;

	PROCEDURE SetCursor(x,y:CARDINAL);
	BEGIN
		Window.GotoXY(x,y);
		IF Insert THEN
			Window.FullCursor
		ELSE
			Window.StdCursor
		END;
	END SetCursor;
	
    PROCEDURE InputAt(VAR String:ARRAY OF CHAR;
                      cposflag:BOOLEAN;cpos:CARDINAL;
                      XPos,YPos:CARDINAL;
                      maxlen:CARDINAL;
                      ClearOnFirstKey:BOOLEAN):BOOLEAN;
    (* ------------------------------------------------------------------ *)
    (* allgemeine String-Input-Routine                                    *)
    (* ------------------------------------------------------------------ *)
    (* PARAMETER:                                                         *)
    (*     NAME          BESCHREIBUNG:                                    *)
    (*     String        I/O: String                                      *)
    (*     cposflag      wenn TRUE, Cursor zu Beginn an Pos. cpos setzen  *)
    (*                   wenn FALSE, Cursor zu Beginn an Pos. 0           *)
    (*     XPos,YPos     Absolute Position                                *)
    (*     Fore,Back     Farbe                                            *)
    (*     maxlen        Maximale String-Lnge                            *)
    (*     ClearOnFirstKey: wenn TRUE, dann wird, falls die erste         *)
    (*                      gedrckte Taste keine Steuertaste war, der    *)
    (*                      String gelscht.                              *)
    (*                                                                    *)
    (* RETURNWERT:                                                        *)
    (*                   TRUE=Eingabe ordnungsgem beendet               *)
    (*                   FALSE=Eingabe mit ESC abgebrochen -> der         *)
    (*                         String wird nicht gendert                 *)
    (* ------------------------------------------------------------------ *)
    (* HISTORISCHE INFORMATION:                                           *)
    (*     NDERUNGEN:                                                    *)
    (* ------------------------------------------------------------------ *)
    (* ERGNZUNGEN:                                                       *)
    (*     keine Koordinaten-berprfung                                  *)
    (*     sehr komfortabel:                                              *)
    (*     * left/right   =   Cursor left/right                           *)
    (*     * CTRL-Left    =   ein Wort nach links                         *)
    (*     * CTRL-Right   =   ein Wort nach rechts                        *)
    (*     * pos1         =   zum Anfang                                  *)
    (*     * ende         =   zum Ende                                    *)
    (*     * insert       =   Toggle Insert/Overwrite                     *)
    (*     * CTRL-Y       =   Zeile lschen                               *)
    (*     * ESC          =   Eingabe abbrechen                           *)
    (* ------------------------------------------------------------------ *)

    VAR i:CARDINAL;                                     (* allgem. Zhler *)
        pos:CARDINAL;                          (* aktuelle Cursorposition *)
        len:CARDINAL;                            (* aktuelle String-Lnge *)
        k:KeyType;                                          (* Tastencode *)
        endflag:BOOLEAN;                               (* Eingabe beendet *)
        str:LONGSTR;           (* Zwischenspeicher - in ihm wird editiert *)
        RedrawAll:BOOLEAN;                  (* den ganzen String ausgeben *)
    BEGIN

		ClearPushBuffer;

		IF maxlen=0 THEN maxlen:=HIGH(String) END;
		
        len:=0;
        WHILE String[len] # eos DO str[len]:=String[len]; INC(len) END;
        str[len]:=eos;  (* str ist der Hilfsstring, in dem editiert wird. *)
                       (* 'String' wird erst nach dem Editieren gendert. *)
                    (* Bei Abbruch bleibt der alte Eingabestring erhalten *)

        IF cposflag THEN                                 (* Cursor setzen *)
            pos:=cpos
        ELSE
            pos:=len;
        END;


        endflag:=FALSE;
        RedrawAll:=TRUE;
        First:=TRUE;

        REPEAT                                           (* Hauptschleife *)

            IF RedrawAll THEN                     (* String ganz ausgeben *)
                WriteStringXY(str,maxlen,XPos,YPos);
                RedrawAll:=FALSE;
            END;

			SetCursor(XPos+pos,YPos);

            k:=RdKeyExt();

            IF k.level=normal THEN
                IF First & ClearOnFirstKey THEN
                    str[0]:=k.key;
                    str[1]:=eos;
                    RedrawAll:=TRUE;
                    pos:=1;
                    len:=1
                ELSE
                    IF Insert THEN
                        IF (pos<maxlen) & (len<maxlen) THEN
                            IF pos=len THEN                (* hinten anhngen *)
								Window.GotoXY(XPos+pos,YPos);
								IO.WrChar(k.key);
                                str[pos]:=k.key;
                                INC(pos);
                                INC(len);
                                str[len]:=eos;
                            ELSE             (* einfgen und Rest verschieben *)
                                FOR i:=len+1 TO pos+1 BY -1 DO                            (* einfgen *)
                                 str[i]:=str[i-1]
                                END;
                                str[pos]:=k.key;
                                INC(pos);
                                INC(len);
                                RedrawAll:=TRUE;
                            END
                        END
                    ELSE                                         (* overwrite *)
                        IF pos<maxlen THEN
							Window.GotoXY(XPos+pos,YPos);
							IO.WrChar(k.key);
                            str[pos]:=k.key;
                            INC(pos);
                            IF pos>=len THEN
                                len:=pos; str[len]:=eos
                            END
                        END
                    END
                END
            ELSE                                 (* Sondertasten-Handling *)
                CASE k.fkey OF

                  delete: IF pos>0 THEN                (* Zeichen lschen *)
                              DEC(pos);
                              DEC(len);
                              IF pos#len THEN   (* Rest nachrcken lassen *)
                                  FOR i:=pos TO len DO str[i]:=str[i+1] END;
                                  RedrawAll:=TRUE
                              ELSE
                                  str[len]:=eos;
								  Window.GotoXY(XPos+pos,YPos);
								  IO.WrStr(" ");
                              END
                          END;

                | left:   IF pos>0 THEN                 (* Cursor bewegen *)
                              DEC(pos);
                              IF (pos+1=len) THEN
                                  IF str[pos]=' ' THEN  (* Blanks am Ende *)
                                      DEC(len);                (* lschen *)
                                      str[len]:=eos
                                  END
                              END
                          END;

                | right:  IF pos<maxlen THEN
                              IF pos=len THEN        (* Bei Bedarf hinten *)
                                  str[pos]:=' ';       (* Blanks anhngen *)
                                  INC(len);
                                  str[len]:=eos
                              END;
                              INC(pos)
                          END;

                | entf:   IF pos<len THEN     (* Zeichen unter dem Cursor *)
                                                               (* lschen *)
                           IF pos=len-1 THEN
                              str[len-1]:=eos;
							  Window.GotoXY(XPos+pos,YPos);
							  IO.WrStr(" ");
                           ELSE
                               FOR i:=pos TO len-1 DO       (* nach links *)
                                   str[i]:=str[i+1]         (* nachrcken *)
                               END;
                               RedrawAll:=TRUE
                           END;
                           DEC(len)
                          END;

                | ende:                         (* Sprung zum String-Ende *)
                          WHILE (len>=1) & (str[len-1]=' ') DO
                              DEC(len)        (* Blanks am Ende entfernen *)
                          END;
                          str[len]:=eos;

                          pos:=len;

                | pos1:   pos:=0;                    (* Sprung zum Anfang *)


                | esc:	endflag:=TRUE;               (* Eingabe abbrechen *)
						IF First THEN PushKey(k); END;

                | return:	endflag:=TRUE;             (* Eingabe beenden *)
							IF MerkReturn THEN
								PushKey(k)
							END;

                | einf:   Insert:=NOT Insert;     (* Toggle Insert on/off *)

                | CtrlY:                         (* Gesamte Zeile lschen *)
                          str[0]:=eos; pos:=0; len:=0; RedrawAll:=TRUE;

                | left2:                           (* ein Wort nach links *)

                          IF pos=len THEN         (* evtl. Blanks am Ende *)
                                                   (* des Strings lschen *)
                              WHILE (pos>0) & (str[pos-1]=' ') DO
                                  DEC(pos);
                                  DEC(len)
                              END;
                          END;

                          IF pos>0 THEN
                              IF (str[pos] IN Alpha) &
                                 (str[pos-1] IN Alpha) THEN
                                  WHILE (pos>0) & (str[pos-1] IN Alpha) DO
                                      DEC(pos)
                                  END
                               ELSE
                                   WHILE (pos>0) & (str[pos] IN Alpha) DO
                                       DEC(pos)
                                   END;
                                   WHILE (pos>0) &
                                         NOT (str[pos] IN Alpha) DO
                                       DEC(pos)
                                   END;
                                   WHILE (pos>0) &
                                         (str[pos-1] IN Alpha) DO
                                       DEC(pos)
                                   END;
                              END;
                          END;

                | right2:
                          WHILE (pos<len) & (str[pos] IN Alpha) DO
                              INC(pos)
                          END;
                          WHILE (pos<len) & NOT (str[pos] IN Alpha) DO
                              INC(pos)
                          END;

				| up,down,tab,tab2,pageup,pagedown,F1..SF10: 	
						endflag:=TRUE;
						PushKey(k);

                END (* CASE *)
            END (* IF *);
            IF ~endflag THEN First:=FALSE END;
        UNTIL endflag;

        Window.StdCursor;
		Window.CursorOff;

		IF (k.level=function) & (k.fkey#esc) THEN
			i:=Str.Length(str);
			WHILE (i>0) & (str[i-1]=' ') DO
				DEC(i)
			END;
			str[i]:=0C;
			Str.Copy(String,str);		   (* Eingabe OK -> String bernehmen *)
			RETURN TRUE;
		END;
		RETURN FALSE
    END InputAt;

BEGIN
	KeyPushed:=FALSE;
	MerkReturn:=FALSE;
	Insert:=TRUE;
END Input.
