(*
  Menu - configurable pull-down menus with callbacks

  Copyright (c) 1991-1993 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE Menu;

FROM Storage IMPORT ALLOCATE,DEALLOCATE;

IMPORT Str,IO,Input;

TYPE MenuType=POINTER TO MenuRec;
     MenuRec=RECORD
                 XPos,YPos:CARDINAL;
                 Width:CARDINAL;
                 AutoHeight:BOOLEAN;
                 Height:CARDINAL; (* when AutoHeight=TRUE automatically filled in *)
				 ResetOnEsc:BOOLEAN;
                 Head,Tail:MenuEntry;
                 CurrentStart,CurrentPos:CARDINAL;
                 CurrentStartElement,CurrentElement:MenuEntry;
                 LastStart,LastPos:CARDINAL;
                 LastStartElement,LastElement:MenuEntry;
                 MenuLength:CARDINAL;
                 IsDisplayed:BOOLEAN;
                 MenuWindow:Window.WinType;
                 MenuTitle:TitleString;
                 WasOpen:BOOLEAN;
				 NeedsRedraw:BOOLEAN;
				 ParentProcess:CARDINAL;
             END;

VAR EventCauseKey:EventCause;

PROCEDURE NumberToString(n:MyNumber;VAR s:ARRAY OF CHAR);
VAR dummy:BOOLEAN;
BEGIN
    Str.CardToStr(LONGCARD(n),s,10,dummy);
END NumberToString;

PROCEDURE StringToNumber(s:ARRAY OF CHAR;VAR n:MyNumber):BOOLEAN;
VAR dummy:BOOLEAN;
    c:CARDINAL;
BEGIN
    c:=CARDINAL(Str.StrToCard(s,10,dummy));
    IF dummy THEN n:=c END;
    RETURN dummy
END StringToNumber;

PROCEDURE NewMenu():MenuType;
VAR m:MenuType;
BEGIN
 NEW(m);
 NEW(m^.Head);
 NEW(m^.Tail);
 m^.Head^.Pred:=NIL;
 m^.Head^.Succ:=m^.Tail;
 m^.Tail^.Succ:=NIL;
 m^.Tail^.Pred:=m^.Head;

 m^.XPos:=DefaultXPos;
 m^.YPos:=DefaultYPos;
 m^.Width:=DefaultWidth;
 m^.AutoHeight:=TRUE;
 m^.IsDisplayed:=FALSE;
 m^.ResetOnEsc:=FALSE;
 m^.CurrentStart:=0;
 m^.CurrentPos:=0;
 m^.CurrentStartElement:=NIL;
 m^.CurrentElement:=NIL;
 m^.MenuLength:=0;
 m^.MenuTitle:="";
 m^.WasOpen:=FALSE;
 m^.NeedsRedraw:=TRUE;
 m^.ParentProcess:=0;
 RETURN(m);
END NewMenu;

PROCEDURE DestroyMenu(VAR m:MenuType; dispose:DisposeAddPtrProc);
VAR pos,temp:MenuEntry;
BEGIN
 IF m # NIL THEN
  pos:=m^.Head;
  WHILE pos # NIL DO
   IF (pos#m^.Tail) & (pos#m^.Head) THEN
    IF (pos^.Info.ActivateSubMenu) THEN
     DestroyMenu(pos^.Info.SubMenu,dispose)
    END;
    IF (pos^.Info.Type=Pointer) THEN
    	dispose(pos^.Info.AddPtr)
    END;
   END;
   temp:=pos;
   pos:=pos^.Succ;
   DISPOSE(temp)
  END;
  DISPOSE(m);
 END;
END DestroyMenu;

PROCEDURE MenuIsEmpty(m:MenuType):BOOLEAN;
BEGIN
	RETURN m^.Head^.Succ=m^.Tail;
END MenuIsEmpty;

PROCEDURE SetTitle(m:MenuType; t:TitleString);
BEGIN
    m^.MenuTitle:=t;
END SetTitle;

PROCEDURE SetOrigin(m:MenuType; x,y:CARDINAL);
BEGIN
    m^.XPos:=x; m^.YPos:=y;
END SetOrigin;

PROCEDURE SetOriginX(m:MenuType; x:CARDINAL);
BEGIN
    m^.XPos:=x;
END SetOriginX;

PROCEDURE SetOriginY(m:MenuType; y:CARDINAL);
BEGIN
    m^.YPos:=y;
END SetOriginY;

PROCEDURE SetWidth(m:MenuType; w:CARDINAL);
BEGIN
    m^.Width:=w;
END SetWidth;

PROCEDURE SetAutoHeight(m:MenuType; flag:BOOLEAN);
BEGIN
    m^.AutoHeight:=flag;
END SetAutoHeight;

PROCEDURE SetHeight(m:MenuType; h:CARDINAL);
BEGIN
    m^.Height:=h;
END SetHeight;

PROCEDURE SetResetOnEsc(m:MenuType; flag:BOOLEAN);
BEGIN
    m^.ResetOnEsc:=flag;
END SetResetOnEsc;

PROCEDURE AddEntry(m:MenuType; e:InfoRecord);
VAR me:MenuEntry;
BEGIN
    NEW(me);
    me^.Info:=e;
    me^.Succ:=m^.Tail;
    me^.Pred:=m^.Tail^.Pred;
    me^.Pred^.Succ:=me;
    m^.Tail^.Pred:=me;
    INC(m^.MenuLength);
END AddEntry;

PROCEDURE AddEntryLastButOne(m:MenuType; e:InfoRecord);
VAR me:MenuEntry;
BEGIN
	IF MenuIsEmpty(m) THEN
		AddEntry(m,e)
	ELSE
	    NEW(me);
   		me^.Info:=e;
    	me^.Succ:=m^.Tail^.Pred;
    	me^.Pred:=m^.Tail^.Pred^.Pred;
    	me^.Pred^.Succ:=me;
    	m^.Tail^.Pred^.Pred:=me;
    	INC(m^.MenuLength);
    END;
END AddEntryLastButOne;

PROCEDURE FixMenuIntegrityAfterInsertLastButOne(m:MenuType);
BEGIN
	IF m^.IsDisplayed THEN
		IF m^.MenuLength=2 THEN
			m^.CurrentStartElement:=m^.CurrentStartElement^.Pred
		END;
		IF m^.CurrentPos+2=m^.MenuLength THEN
			INC(m^.CurrentPos);
		END;
	ELSE
		m^.CurrentStartElement:=NIL
	END;
END FixMenuIntegrityAfterInsertLastButOne;

PROCEDURE AddEntrySorted(m:MenuType; e:InfoRecord);
VAR me,tmp:MenuEntry;
BEGIN
    NEW(me);
    me^.Info:=e;
	tmp:=m^.Head^.Succ;
	WHILE (tmp#m^.Tail) & (Str.Compare(tmp^.Info.Text,e.Text)<0) DO
		tmp:=tmp^.Succ
	END;
    me^.Succ:=tmp;
    me^.Pred:=tmp^.Pred;
    me^.Pred^.Succ:=me;
    tmp^.Pred:=me;
    INC(m^.MenuLength);
END AddEntrySorted;

PROCEDURE DestroyEntry(m:MenuType; e:MenuEntry; dispose:DisposeAddPtrProc);
BEGIN
	e^.Pred^.Succ:=e^.Succ;
	e^.Succ^.Pred:=e^.Pred;
	IF e=m^.CurrentStartElement THEN
		m^.CurrentStartElement:=m^.CurrentStartElement^.Succ
	END;
	IF (e^.Info.Type=Pointer) THEN
		dispose(e^.Info.AddPtr)
	END;
	m^.CurrentElement:=m^.CurrentElement^.Succ;
	DEC(m^.MenuLength);
	SetRedrawFlag(m);
	DISPOSE(e);
END DestroyEntry;

PROCEDURE FormatToString(VAR e:InfoRecord; VAR str:ItemString; width:CARDINAL);
VAR temp:ARRAY[0..20] OF CHAR;
    i:CARDINAL;
BEGIN
    IF e.TagAble THEN
       str:=' [ ] ';
       IF e.Tagged THEN
          str[2]:='X';
       END;
       Str.Concat(str,str,e.Text);
    ELSE
        str:=" "; Str.Concat(str,str,e.Text);
    END;
    IF (e.Type=ChangeNumber) OR (e.Type=ConstNumber) THEN
       NumberToString(e.Num,temp);
       Str.Concat(str,str,' : ');
       e.InternalNumberPos:=Str.Length(str);
       Str.Concat(str,str,temp);
    END;
    IF (e.Type=String) THEN
       Str.Concat(str,str,' : ');
       e.InternalStringPos:=Str.Length(str);
       Str.Concat(str,str,e.string);
    END;
    i:=Str.Length(str);
    WHILE (i<width+2) DO
        str[i]:=' ';
        INC(i);
    END;
    str[width+2]:=0C;
END FormatToString;

PROCEDURE InputNumber(m:MenuType; VAR e:InfoRecord);
VAR temp:ARRAY[0..10] OF CHAR;
    c,maxlen:CARDINAL;

BEGIN
    IF e.Type=ChangeNumber THEN

       maxlen:=5;
       IF (e.InternalNumberPos+5>=m^.Width+1) THEN
          maxlen:=m^.Width+1-e.InternalNumberPos
       END;

       Window.Use(m^.MenuWindow);
       Window.TextColor(DefaultInputFore);
       Window.TextBackground(DefaultInputBack);

       NumberToString(e.Num,temp);
       IF Input.InputAt(temp,FALSE,0,
                        e.InternalNumberPos+1,
                        m^.CurrentPos-m^.CurrentStart+1,
                        maxlen,TRUE) THEN
         IF StringToNumber(temp,e.Num) THEN END;
         SetRedrawFlag(m);
       END;
    END;
END InputNumber;

PROCEDURE InputString(m:MenuType; VAR e:InfoRecord);
VAR maxlen:CARDINAL;

BEGIN
    IF e.Type=String THEN

       maxlen:=m^.Width+1-e.InternalStringPos;
       IF maxlen>80 THEN maxlen:=80 END;

       Window.Use(m^.MenuWindow);
       Window.TextColor(DefaultInputFore);
       Window.TextBackground(DefaultInputBack);

       IF Input.InputAt(e.string,FALSE,0,
                        e.InternalStringPos+1,
                        m^.CurrentPos-m^.CurrentStart+1,
                        maxlen,TRUE) THEN
         SetRedrawFlag(m);
       END;
    END;
END InputString;


PROCEDURE WriteMenu(m:MenuType);
VAR i,j:CARDINAL;
    str:ItemString;
    me:MenuEntry;
BEGIN
    Window.Use(m^.MenuWindow);
	Window.TextColor(DefaultMenuText);
    Window.TextBackground(DefaultBackground);

	WHILE m^.CurrentStart+m^.Height<=m^.CurrentPos DO
		INC(m^.CurrentStart);
		m^.CurrentStartElement:=m^.CurrentStartElement^.Succ;
	END;

	WHILE m^.CurrentStart>m^.CurrentPos DO
		DEC(m^.CurrentStart);
		m^.CurrentStartElement:=m^.CurrentStartElement^.Pred;
	END;

    IF (~m^.WasOpen) OR (m^.NeedsRedraw) OR
    	( (m^.CurrentStart#m^.LastStart+1) &
    	  (m^.CurrentStart+1#m^.LastStart) &
    	  (m^.CurrentStart#m^.LastStart)  ) THEN
                                                    (* Men ganz neu aufbauen *)
        m^.WasOpen:=TRUE;
		m^.NeedsRedraw:=FALSE;
        Window.Clear;
        i:=m^.CurrentStart;
        me:=m^.CurrentStartElement;
        j:=1;
        WHILE (i<m^.MenuLength) & (j<=m^.Height) DO
            FormatToString(me^.Info,str,m^.Width);
            Window.GotoXY(1,j);
            IF i=m^.CurrentPos THEN
               Window.TextColor(DefaultCursorFore);
               Window.TextBackground(DefaultCursorBack);
               IO.WrStr(str);
               Window.TextColor(DefaultMenuText);
               Window.TextBackground(DefaultBackground);
            ELSE
               IO.WrStr(str);
            END;
            INC(i);
            INC(j);
            me:=me^.Succ;
        END;
    ELSE

        (* undisplay cursor-line *)
        Window.GotoXY(1,1+m^.LastPos-m^.LastStart);
        FormatToString(m^.LastElement^.Info,str,m^.Width);
        IO.WrStr(str);

        IF m^.LastStart=m^.CurrentStart THEN        (* no scrolling is needed *)
            Window.GotoXY(1,1+m^.CurrentPos-m^.CurrentStart);
        ELSE                                           (* scrolling is needed *)
            IF (m^.CurrentStart=m^.LastStart+1) THEN (*Scroll up*)
                Window.GotoXY(1,1);
                Window.DelLine;
                Window.GotoXY(1,m^.Height);
            ELSIF (m^.CurrentStart+1=m^.LastStart) THEN (*Scroll down*)
                Window.GotoXY(1,1);
                Window.InsLine;
                Window.GotoXY(1,1);
            END;
        END;

        FormatToString(m^.CurrentElement^.Info,str,m^.Width);
        Window.TextColor(DefaultCursorFore);
        Window.TextBackground(DefaultCursorBack);
        IO.WrStr(str);
    END;
    m^.LastPos:=m^.CurrentPos;
    m^.LastStart:=m^.CurrentStart;
    m^.LastElement:=m^.CurrentElement;
    m^.LastStartElement:=m^.CurrentStartElement;
END WriteMenu;

PROCEDURE DisplayMenu(m:MenuType);
VAR wd:Window.WinDef;
BEGIN
    IF m^.CurrentStartElement=NIL THEN
       m^.CurrentStartElement:=m^.Head^.Succ;
       m^.CurrentElement:=m^.Head^.Succ;
	   m^.CurrentStart:=0;
	   m^.CurrentPos:=0;
    END;
    IF ~m^.IsDisplayed THEN
       IF m^.AutoHeight THEN
           m^.Height:=m^.MenuLength;
           IF m^.YPos+m^.Height+1>MaxYPos-1 THEN
               m^.Height:=MaxYPos-m^.YPos-2
           END;
       END;
       wd.X1:=m^.XPos;
       wd.Y1:=m^.YPos;
       wd.X2:=m^.XPos+m^.Width+3;
       wd.Y2:=m^.YPos+m^.Height+1;
       wd.Foreground:=DefaultMenuText;
       wd.Background:=DefaultBackground;
       wd.CursorOn:=FALSE;
       wd.WrapOn:=FALSE;
       wd.Hidden:=FALSE;
       wd.FrameOn:=TRUE;
       wd.FrameDef:=ActiveMenuFrame;
       wd.FrameFore:=MenuBorder;
       wd.FrameBack:=DefaultBackground;
       m^.MenuWindow:=Window.Open(wd);
       Window.SetTitle(m^.MenuWindow,m^.MenuTitle,Window.CenterUpperTitle);
       m^.IsDisplayed:=TRUE;
	ELSE
		Window.PutOnTop(m^.MenuWindow);
    END;
	WriteMenu(m);
END DisplayMenu;

PROCEDURE UnDisplayMenu(m:MenuType);
BEGIN
    m^.IsDisplayed:=FALSE;
    m^.WasOpen:=FALSE;
    Window.Close(m^.MenuWindow);
END UnDisplayMenu;

PROCEDURE SetActiveFrame(m:MenuType);
BEGIN
	Window.SetFrame(m^.MenuWindow,ActiveMenuFrame,MenuBorder,
					DefaultBackground);
END SetActiveFrame;

PROCEDURE SetInactiveFrame(m:MenuType);
BEGIN
	Window.SetFrame(m^.MenuWindow,InactiveMenuFrame,MenuBorder,
					DefaultBackground);
END SetInactiveFrame;

PROCEDURE UseMenu(m:MenuType):MenuEntry;
VAR k:Input.KeyType;
    i:CARDINAL;
	EndFlag,TempFlag:BOOLEAN;
	ReturnValue,CE:MenuEntry;
BEGIN
    DisplayMenu(m);
	SetActiveFrame(m);

	EndFlag:=FALSE;
	
    REPEAT
         k:=Input.RdKeyExt();
         IF k.level=Input.function THEN
            CASE k.fkey OF
              Input.up: IF (m^.CurrentPos>0) THEN
                           DEC(m^.CurrentPos);
                           m^.CurrentElement:=m^.CurrentElement^.Pred;
                           DisplayMenu(m);
                        END;
            | Input.down: IF (m^.CurrentPos<m^.MenuLength-1) THEN
                           INC(m^.CurrentPos);
                           m^.CurrentElement:=m^.CurrentElement^.Succ;
                           DisplayMenu(m);
                          END;

            | Input.pageup: i:=m^.Height-1;
            				WHILE (i>0) & (m^.CurrentPos>0) DO
                           		DEC(m^.CurrentPos);
                           		m^.CurrentElement:=m^.CurrentElement^.Pred;
                           		DEC(i);
                           	END;
                            DisplayMenu(m);
            | Input.pagedown: 	i:=m^.Height-1;
            					WHILE (i>0) & (m^.CurrentPos<m^.MenuLength-1) DO
                           			INC(m^.CurrentPos);
                           			m^.CurrentElement:=m^.CurrentElement^.Succ;
                           			DEC(i)
                           		END;
                           		DisplayMenu(m);
            | Input.pos1: 	m^.CurrentPos:=0;
                           	m^.CurrentElement:=m^.Head^.Succ;
                            DisplayMenu(m);
            | Input.ende:	m^.CurrentPos:=m^.MenuLength-1;
            				m^.CurrentElement:=m^.Tail^.Pred;
                           	DisplayMenu(m);
			| Input.return:
                        EventCauseKey:=Return_Key;
						CE:=m^.CurrentElement;
						IF (CE^.Info.ActivateSubMenu) THEN
							TempFlag:=TRUE;
    						IF CE^.Info.CallEvent THEN
								SetInactiveFrame(m);
    							TempFlag:=DispatchEvent(m,CE);
								SetActiveFrame(m);
    						END;
							IF TempFlag THEN
								SetInactiveFrame(m);
								ReturnValue:=UseMenu(CE^.Info.SubMenu);
								SetActiveFrame(m);
								IF ReturnValue=NIL THEN
									Window.PutOnTop(m^.MenuWindow);
								ELSE
									EndFlag:=CE^.Info.LeaveOnReturn
								END;
							END;
						ELSE
							ReturnValue:=CE;
    						IF CE^.Info.CallEvent THEN
								SetInactiveFrame(m);
    							IF DispatchEvent(m,CE) THEN
									EndFlag:=CE^.Info.LeaveOnReturn;
								END;
								SetActiveFrame(m);
							ELSE
								EndFlag:=CE^.Info.LeaveOnReturn;
    						END;
						END;
						DisplayMenu(m);
			| Input.esc:
						SetInactiveFrame(m);
						ReturnValue:=NIL;
						EndFlag:=TRUE;
						IF  m^.ResetOnEsc THEN
							m^.CurrentStartElement:=NIL;
							m^.WasOpen:=FALSE;
						END;
            END;
		ELSE
			IF (k.key=' ') & (m^.CurrentElement^.Info.TagAble) THEN
                EventCauseKey:=Tag_Key;
				WITH m^.CurrentElement^.Info DO
					Tagged:=~Tagged
				END;
				IF m^.CurrentElement^.Info.CallEvent THEN
					SetInactiveFrame(m);
					IF DispatchEvent(m,m^.CurrentElement) THEN END;
					SetActiveFrame(m);
				END;
				DisplayMenu(m);
			END;				
        END;
    UNTIL EndFlag;

	UnDisplayMenu(m);

	RETURN ReturnValue;
END UseMenu;

PROCEDURE SetParentProcess(m:MenuType;p:CARDINAL);
BEGIN
	m^.ParentProcess:=p;
END SetParentProcess;

PROCEDURE GetParentProcess(m:MenuType):CARDINAL;
BEGIN
	RETURN m^.ParentProcess;
END GetParentProcess;

PROCEDURE SetRedrawFlag(m:MenuType);
BEGIN
	m^.NeedsRedraw:=TRUE;
END SetRedrawFlag;

PROCEDURE DoNothingWithEvent(m:MenuType;e:MenuEntry):BOOLEAN;
BEGIN
	RETURN TRUE;
END DoNothingWithEvent;

PROCEDURE GetFirstEntry(m:MenuType):MenuEntry;
VAR e:MenuEntry;
BEGIN
	e:=m^.Head^.Succ;
	IF e=m^.Tail THEN
		e:=NIL
	END;
	RETURN e;
END GetFirstEntry;

PROCEDURE GetNextEntry(m:MenuType;e:MenuEntry):MenuEntry;
BEGIN
	e:=e^.Succ;
	IF e=m^.Tail THEN
		e:=NIL
	END;
	RETURN e;
END GetNextEntry;

PROCEDURE GetMenuLength(m:MenuType):CARDINAL;
BEGIN
    IF m=NIL THEN RETURN 0 END;
	RETURN m^.MenuLength;
END GetMenuLength;

PROCEDURE GetEventCause():EventCause;
BEGIN
    RETURN EventCauseKey
END GetEventCause;

PROCEDURE SetColor;
BEGIN
    DefaultBackground:=Window.Black;
    DefaultMenuText:=Window.LightCyan;
    DefaultCursorBack:=Window.Blue;
    DefaultCursorFore:=Window.Yellow;
    DefaultInputBack:=Window.Cyan;
    DefaultInputFore:=Window.White;
END SetColor;

PROCEDURE SetMono;
BEGIN
    DefaultBackground:=Window.Black;
    DefaultMenuText:=Window.LightGray;
    DefaultCursorBack:=Window.LightGray;
    DefaultCursorFore:=Window.Black;
    DefaultInputBack:=Window.LightGray;
    DefaultInputFore:=Window.White;
END SetMono;

BEGIN
	DispatchEvent:=DoNothingWithEvent;
    SetColor;
END Menu.
