(*
  MParser - read menu configuration from file

  Copyright (c) 1991-1993 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE MParser;

IMPORT FIO,IO,Str,Lib;

(*
Aufbau eines Men-Eintrages:
% 1. Name des Eintrages
% 2. Code: ABCDE A=0/1: TagAble;
%                B=0/1: number value follows
%                C=0/1: Const/Var Number or string (if B=0)
%                D=0/1: Call Event-Procedure with Event No. given
%                E=0/1: Exit menu on Return
% (3. Nummer)
% (4. Event-Nummer)
% Example:
% Entry1  |01011|123.45|125
%
% Steuerzeichen:
% $Wzahl   : Width = zahl
% $A0/1    : AutoHeight = 0/1         default: 1
% $Hzahl   : Hhe = zahl
% $Xzahl   : XPos=Zahl                default: DefaultXPos
% $Yzahl   : YPos=Zahl                default: DefaultYPos
% $Tstring : Title = string
% $R0/1    : Reset position on ESC    default: 0
*)


VAR eof:BOOLEAN;
	f:FIO.File;

	FileBuffer:ARRAY[0..512*8+FIO.BufferOverhead] OF BYTE;

VAR line,text:ARRAY[0..100] OF CHAR;
    string:Menu.ItemString;
	tmp:ARRAY[0..80] OF CHAR;
	Flag:ARRAY[0..4] OF BOOLEAN;
	num:CARDINAL;
	EventNum:CARDINAL;
		
PROCEDURE NextLine;
VAR i,j:CARDINAL;
BEGIN
	REPEAT
		FIO.RdStr(f,line);
	UNTIL ((line[0]#'%') & (line[0]#0C)) OR (FIO.EOF);
	i:=0;
	WHILE (line[i]=CHAR(9)) OR (line[i]=' ') DO INC(i) END;
	IF i>0 THEN
		j:=0;
		REPEAT
			line[j]:=line[i];
			IF line[j]='~' THEN line[j]:=' ' END;
			INC(i);
			INC(j);
		UNTIL line[i-1]=0C;
	END;
END NextLine;

PROCEDURE GetParameter():CARDINAL;
VAR tmp:ARRAY[0..20] OF CHAR;
	ok:BOOLEAN;
BEGIN
	Str.Slice(tmp,text,2,20);
	RETURN CARDINAL(Str.StrToInt(tmp,10,ok));
END GetParameter;

PROCEDURE ParseLine;
VAR i,j,item:CARDINAL;
	tmp:ARRAY[0..80] OF CHAR;
	ok:BOOLEAN;
BEGIN
    IF line[0]='$' THEN
	text:=line;
    ELSE
    	item:=2;
    	Str.ItemS(text,line,"|",0);
    	Str.ItemS(tmp,line,"|",1);
    	FOR i:=0 TO 4 DO
    		Flag[i]:=tmp[i]='1'
    	END;
    	IF Flag[1] THEN                                     (* Number follows *)
    	   Str.ItemS(tmp,line,"|",item);
    	   INC(item);
	   num:=CARDINAL(Str.StrToCard(tmp,10,ok));
          ELSIF Flag[2] THEN
    	   Str.ItemS(string,line,"|",item);
    	   INC(item);
           IF (string[0]="") & (string[1]=0C) THEN string[0]:=0C END;
	END;
	IF Flag[3] THEN                                 (* Call Event-Handler *)
    	   Str.ItemS(tmp,line,"|",item);
	   EventNum:=CARDINAL(Str.StrToCard(tmp,10,ok));
    	END;
    END;
END ParseLine;

PROCEDURE OpenFile(name:ARRAY OF CHAR);
BEGIN
	f:=FIO.Open(name);
	FIO.AssignBuffer(f,FileBuffer);
	NextLine;
END OpenFile;

PROCEDURE CloseFile();
BEGIN
	FIO.Close(f);
END CloseFile;

PROCEDURE Parse(VAR m:Menu.MenuType;init:InitEntryEvent);
VAR	info:Menu.InfoRecord;
	mtit:Menu.TitleString;
BEGIN
	IF line[0]#'{' THEN IO.WrStr("kein Block-Start!!!"); HALT END;
	NextLine;
	m:=Menu.NewMenu();
	WHILE line[0]#'}' DO
		ParseLine;
		NextLine;
		IF text[0]='$' THEN
			CASE text[1] OF
			  'X': Menu.SetOriginX(m,GetParameter());
			| 'Y': Menu.SetOriginY(m,GetParameter());
			| 'W': Menu.SetWidth(m,GetParameter());
			| 'H': Menu.SetHeight(m,GetParameter());
			| 'A': Menu.SetAutoHeight(m,BOOLEAN(GetParameter()));
			| 'R': Menu.SetResetOnEsc(m,BOOLEAN(GetParameter()));
			| 'T':	Str.Slice(mtit,text,2,80);
					Menu.SetTitle(m,mtit);
			ELSE
				Lib.FatalError(text);
			END;
		ELSE
			Str.Copy(info.Text,text);
			info.TagAble:=Flag[0];
			IF info.TagAble THEN
				info.Tagged:=FALSE;
			END;
			IF Flag[1] THEN
				IF Flag[2] THEN
					info.Type:=Menu.ChangeNumber;
				ELSE
					info.Type:=Menu.ConstNumber;
				END;
				info.Num:=num;
			ELSIF Flag[2] THEN
			          info.Type:=Menu.String;
			          info.string:=string;
			ELSE
				info.Type:=Menu.NoInfo
			END;
			info.CallEvent:=Flag[3];
			IF Flag[3] THEN
				info.EventNo:=EventNum;
			END;
			info.LeaveOnReturn:=Flag[4];
			IF line[0]='{' THEN
				info.ActivateSubMenu:=TRUE;
				Parse(info.SubMenu,init);
			ELSE
				info.ActivateSubMenu:=FALSE
			END;
			IF (init#NULLPROC) & (info.CallEvent) THEN
               init(info)
            END;
			Menu.AddEntry(m,info);
		END;			
	END;
	NextLine;
END Parse;

END MParser.
